from enum import Enum


class Typeface(Enum):
    ROMAN = 0
    SANS_SERIF = 1


class Margin(Enum):
    TOP = 0
    RIGHT = 1
    BOTTOM = 2
    LEFT = 3


class PageLengthUnit(Enum):
    LINES = 0
    INCHES = 1


class Justification(Enum):
    LEFT = 0
    CENTER = 1
    RIGHT = 2
    FULL = 3


class CharacterSetVariant(Enum):
    USA = 0
    FRANCE = 1
    GERMANY = 2
    UK = 3
    DENMARK_I = 4
    SWEDEN = 5
    ITALY = 6
    SPAIN_I = 7
    JAPAN = 8
    NORWAY = 9
    DENMARK_II = 10
    SPAIN_II = 11
    LATIN_AMERICA = 12
    KOREA = 13
    LEGAL = 64


class CharacterTable(Enum):
    def __new__(cls, *args, **kwds):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, d2, d3):
        self.d2 = d2
        self.d3 = d3

    def __str__(self):
        return f'{self.name} (d2={self.d2}, d3={self.d3})'

    # int value is d2 and d3 parameters in the manual
    # d1 to be setseparately
    # [C-75]

    ITALIC = 0, 0
    PC437_US = 1, 0
    PC437_GREEK = 1, 16
    PC932_JAPANESE = 2, 0
    PC850_MULTILINGUAL = 3, 0
    PC851_GREEK = 4, 0
    PC853_TURKISH = 5, 0
    PC855_CYRILLIC = 6, 0
    PC860_PORTUGAL = 7, 0
    PC863_CANADA_FRENCH = 8, 0
    PC865_NORWAY = 9, 0
    PC852_EAST_EUROPE = 10, 0
    PC857_TURKISH = 11, 0
    PC862_HEBREW = 12, 0
    PC864_ARABIC = 13, 0
    PC_AR864 = 13, 32
    PC866_RUSSIAN = 14, 0
    BULGARIAN_ASCII = 14, 16
    PC866_LAT_LATVIAN = 14, 32
    PC869_GREEK = 15, 0
    USSR_GOST_RUSSIAN = 16, 0
    ECMA_94_1 = 17, 0
    KU42_KU_THAI = 18, 0
    TIS11_TS_988_THAI = 19, 0
    TIS18_GENERAL_THAI = 20, 0
    TIS17_SIC_STD_T21HAI = 21, 0
    TIS13_IBM_STD_T22HAI = 22, 0
    TIS16_SIC_OLD_THAI = 23, 0
    PC861_ICELAND = 24, 0
    BRASCII = 25, 0
    ABICOMP = 26, 0
    MAZOWIA_POLAND = 27, 0
    CODE_MJK_CSFR = 28, 0
    ISO8859_7_LATIN_GREEK = 29, 7
    ISO8859_1_LATIN_1 = 29, 16
    TSM_WIN_THAI_SYSTEM_MANAGER = 30, 0
    ISO_LATIN_1T_TURKISH = 31, 0
    BULGARIA = 32, 0
    HEBREW_7 = 33, 0
    HEBREW_8 = 34, 0
    ROMAN_ = 35, 0
    PC_LITHUANIA = 36, 0
    ESTONIA_ESTONIA = 37, 0
    ISCII = 38, 0
    PC_ISCII = 39, 0
    PC_APTEC = 40, 0
    PC = 41, 0
    PC720 = 42, 0
    OCR_B = 112, 0
    ISO_LATIN_1 = 127, 1
    ISO_8859_2_LATIN_2 = 127, 2
    ISO_LATIN_7_GREEK = 127, 7,



from .parameters import Margin, PageLengthUnit, Typeface, Justification, CharacterSetVariant, CharacterTable
from .commands import Commands
from .commands_9_pin import Commands_9_Pin
from .commands_24_48_pin import Commands_24_48_Pin
from .lookup import lookup_by_pins
from .exceptions import InvalidEncodingError



from .commands import Commands
from .commands_9_pin import Commands_9_Pin
from .commands_24_48_pin import Commands_24_48_Pin


def lookup_by_pins(pins) -> Commands:
    """Lookup commands by the number of pins."""
    match pins:
        case 9:
            return Commands_9_Pin()
        case 24, 48:
            return Commands_24_48_Pin()
        case _:
            raise ValueError(f'Invalid number of pins: {pins}')



from ..commands import Commands
from ..printer import Printer


def print_and_reset(printers: [Printer], cmd: Commands, reset_sequence=None):
    for printer in printers:
        printer.send(cmd.buffer)

    if reset_sequence:
        reset_sequence()



import sys

from .printer import Printer


class DebugPrinter(Printer):

    def __init__(self, output=sys.stdout):
        self.output = output

    def send(self, sequence: bytes):
        print_out = lambda x: print(x, file=self.output)
        print_out('-' * 72)
        print_out(str(sequence))
        print_out(str(sequence.hex(':', 1)))

    def close(self):
        pass



import sys
import argparse


from ..printer import DebugPrinter, PrinterNotFound, UsbPrinter
from ..commands import lookup_by_pins
from .test_page import print_test_page
from .poem import print_poem
from .char_tables import print_char_table
from .i18n_char_set import print_i18n_char_set


def print_function(demo: str):
    match demo:
        case 'testpage':
            return print_test_page
        case 'poem':
            return print_poem
        case 'chartable':
            return print_char_table
        case 'charset':
            return print_i18n_char_set
        case _:
            raise ValueError(f'Unknown demo: {demo}')


def demo(vendor_id: int, product_id: int, pins: int, print_function):
    # Actual printer
    printer = UsbPrinter(id_vendor=vendor_id, id_product=product_id, log_io=sys.stdout)
    # Debug (shows the commands with formatting)
    debug = DebugPrinter()
    commands = lookup_by_pins(pins)
    print_function([printer, debug], commands)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Print a demo page')
    parser.add_argument(
        'demo', type=str, choices=['testpage', 'poem', 'chartable', 'charset'], help='Binary file to print'
    )
    parser.add_argument('-c', '--connector', type=str, required=True, choices=['usb'], help='Connector type')
    parser.add_argument('-p', '--pins', type=int, required=True, choices=[9, 24, 48], help='Number of pins')
    parser.add_argument('--vendor-id', type=str, required=True, help='USB Vendor ID (eg 0x04b8)')
    parser.add_argument('--product-id', type=str, required=True, help='USB Product ID (eg 0x0005)')
    args = parser.parse_args()
    print_function = print_function(args.demo)

    try:
        vendor_id = int(args.vendor_id, 16)
        product_id = int(args.product_id, 16)
    except ValueError:
        print('Invalid vendor/product ID', file=sys.stderr)
        exit(1)

    try:
        demo(vendor_id, product_id, args.pins, print_function)
    except PrinterNotFound as e:
        print(f'Printer not found: {e}', file=sys.stderr)
        exit(1)



import pytest
from escp import CharacterSetVariant

from test import CommandsDefault


@pytest.fixture
def commands():
    return CommandsDefault()


def test_magic_text_no_magic_needed(commands):
    assert commands.magic_text("Hello world").buffer == b'Hello world'


def test_magic_text_no_magic_needed_special_char(commands):
    assert commands.magic_text("Hello $").buffer == b'Hello $'


def test_magic_text_switch_once(commands):
    assert commands.magic_text("Salut à toi").buffer == b'Salut \x1bR\x01\x40\x1bR\x00 toi'


def test_magic_text_switch_once_for_multiple_characters(commands):
    assert commands.magic_text("Salut ààà toi").buffer == b'Salut \x1bR\x01\x40\x40\x40\x1bR\x00 toi'


def test_magic_text_switch_twice(commands):
    assert commands.magic_text("Salut à toi©").buffer == b'Salut \x1bR\x01\x40\x1bR\x00 toi\x1bR\x40\x7b\x1bR\x00'


def test_magic_text_no_switch_already_good_char_set(commands):
    commands.character_set(CharacterSetVariant.FRANCE)
    assert commands.magic_text("Salut à toi").buffer == b'\x1bR\x01Salut \x40 toi'



from escp import CharacterSetVariant

from ..commands import Commands, Justification, PageLengthUnit, CharacterTable
from ..printer import Printer
from .common import print_and_reset


PAGE_LENGTH_INCHES = 12


def print_i18n_char_set(printers: [Printer], cmd: Commands):
    def _print_and_reset():
        print_and_reset(printers, cmd)

    cmd.init()

    cmd.text('CHARACTER SET').cr_lf()
    cmd.text('=============').cr_lf(3)

    cmd.text('Sample of the available character sets').cr_lf()
    cmd.text('Sequence: ESC R n').cr_lf()
    cmd.text('ref: R-41').cr_lf(2)

    codes = [35, 36, 64, 91, 92, 93, 94, 96, 123, 124, 125, 126]

    cmd.text(9 * ' ').text('Hex  ')
    for code in codes:
        cmd.text(hex(code)[2:]).text(' ' * 2)
    cmd.cr_lf()

    character_sets = [
        CharacterSetVariant.USA, CharacterSetVariant.FRANCE, CharacterSetVariant.GERMANY, CharacterSetVariant.UK
    ]
    for i18n_char_set in range(0, 4):
        cs = character_sets[i18n_char_set]
        cmd.text('{0:<12}'.format(cs.name))
        cmd.character_set(cs)
        for code in codes:
            cmd.text(chr(code)).text(' ' * 3)
        cmd.cr_lf()
    cmd.form_feed()

    _print_and_reset()

    for printer in printers:
        printer.close()



import sys
import typing

import usb.core
import usb.util

from .printer import Printer
from .exceptions import PrinterNotFound


class UsbPrinter(Printer):
    # Reference printer Epson LX-300+II
    # ID_VENDOR = 0x04b8
    # ID_PRODUCT = 0x0005

    device: usb.core.Device | None

    def __init__(
            self,
            *,
            id_vendor: int, id_product: int, endpoint_out=0x01, endpoint_in=0x82, log_io: typing.IO | None = None
    ):
        self.device = None
        self.id_vendor = id_vendor
        self.id_product = id_product
        self.endpoint_out = endpoint_out
        self.endpoint_in = endpoint_in
        self.log_io = log_io

        devices = usb.core.show_devices()
        self.log(devices)

        self.device: usb.core.Device = usb.core.find(idVendor=self.id_vendor, idProduct=self.id_product)
        if not self.device:
            hex_value = lambda x: f'0x{x:04x}'
            raise PrinterNotFound(f'USB id_vendor={hex_value(id_vendor)} id_product={hex_value(id_product)}')
        self.log(self.device)
        # TODO Check is printer

        self.detach_kernel_driver()

        self.log("Reset device")
        self.device.reset()

    def detach_kernel_driver(self):
        module: str = self.device.backend.__module__
        self.log(f"Module: ${module}")
        if module.endswith('libusb1'):
            # TODO Explicit check for errno 13 (permission denied)
            check_driver = None
            try:
                check_driver = self.device.is_kernel_driver_active(interface=0)
            except usb.core.USBError as e:
                self.log(f"Failed to check kernel driver activation: ${e}", file=sys.stderr)

            if check_driver is None or check_driver:
                try:
                    self.device.detach_kernel_driver(0)
                except usb.core.USBError as e:
                    self.log(f"Failed to detach kernel driver: ${e}", file=sys.stderr)

    def send(self, sequence: bytes):
        self.device.write(self.endpoint_out, sequence, 5)

    def close(self):
        self.log('Releasing USB')
        if self.device:
            usb.util.dispose_resources(self.device)
        self.device = None
        self.log_io.flush()

    def log(self, message: str):
        if self.log_io:
            self.log_io.write(message)

    def __del__(self):
        self.close()



from typing import TypeVar
from abc import ABC, abstractmethod

from .exceptions import InvalidEncodingError
from .parameters import Margin, PageLengthUnit, Typeface, Justification, CharacterSetVariant, CharacterTable
from .magic_encoding import magic_encoding


def int_to_bytes(value: int) -> bytes:
    return int.to_bytes(value, length=1, byteorder='big', signed=False)


T = TypeVar('T', bound='Commands')


class Commands(ABC):

    cmds = {
        'init': b'\x1b@',
        'draft': b'\x1bx',
        'cr_lf': b'\x0d\x0a',
        'character_width_10': b'\x1bP',
        'character_width_12': b'\x1bM',
        'character_width_15': b'\x1bg',
        'bold_on': b'\x1bE',
        'bold_off': b'\x1bF',
        'italic_on': b'\x1b4',
        'italic_off': b'\x1b5',
        'double_strike_on': b'\x1bG',
        'double_strike_off': b'\x1bH',
        'underline_on': b'\x1b-\x01',
        'underline_off': b'\x1b-\x00',
        'superscript_on': b'\x1bS\x00',
        'subscript_on': b'\x1bS\x01',
        'superscript_off': b'\x1bT',
        'subscript_off': b'\x1bT',
        'upper_control_codes_printing_on': b'\x1b\x36',
        'upper_control_codes_printing_off': b'\x1b\x37',
        'control_codes_printing': b'\x1b\x49',
        'typeface': b'\x1bk',
        'margin_left': b'\x1bl',
        'margin_right': b'\x1bQ',
        'margin_bottom': b'\x1bN',
        'page_length_in_lines': b'\x1bC',
        'page_length_in_inches': b'\x1bC\x00',
        'double_character_width': b'\x1bW',
        'double_character_width_one_line': b'\x1b\x0e',
        'double_character_height': b'\x1bw',
        'extra_space': b'\x1b ',
        'condensed_on': b'\x1b\x0f',
        'condensed_off': b'\x1b\x12',
        'line_spacing_1_6': b'\x1b2',
        'line_spacing_1_8': b'\x1b0',
        'proportional': b'\x1bp',
        'justify': b'\x1ba',
        'form_feed': b'\x0c',
        'character_set': b'\x1bR',
        'select_character_table': b'\x1b\x74',
        'assign_character_table': b'\x1b\x28\x74',
    }

    _buffer: bytearray

    def __init__(self):
        self.current_character_set = CharacterSetVariant.USA
        self.clear()

    def _commands(self):
        return self.cmds

    def init(self) -> T:
        return self._append_cmd('init')

    def draft(self, enabled: bool) -> T:
        """Change the print quality to draft or letter quality.

        Point sizes 10.5 and 21 only.
        LQ quality for ESC/P2 and ESC/P.
        NLQ for 9-pin printers.
        """
        return self._append_cmd('draft', int_to_bytes(1 if enabled else 0))

    @abstractmethod
    def is_valid_character_table(self, table: int) -> bool:
        """Test if the given character table is valid.

        Depends on the printer number of pins.
        """
        raise NotImplementedError()

    def select_character_table(self, table: int) -> T:
        """Select the character table.

        There are 2 or 4 character tables depending on the printer type [C-77].
        Use the ESC ( t command to assign any registered character table to any character table.

        :param table: Character table number.
        """
        if not self.is_valid_character_table(table):
            raise ValueError(f'Invalid character table {table}')
        return self._append_cmd('select_character_table', int_to_bytes(table))

    def assign_character_table(self, table: int, ct: CharacterTable) -> T:
        """Assign the given character table `ct` to the character `table`.

        :param table: Character table number to assign to.
        :param ct: Character table to assign.
        """
        if not self.is_valid_character_table(table):
            raise ValueError('Invalid character table')
        return self._append_cmd(
            'assign_character_table',
            b'\x03\x00' + int_to_bytes(table) + int_to_bytes(ct.d2) + int_to_bytes(ct.d3)
        )

    def character_set(self, cs: CharacterSetVariant) -> T:
        """Change up to 12 of the characters in the current character table.

        These 12 characters are called international character sets
        because they correspond to characters commonly used in several foreign
        languages. [R-41]

        It is up to the implementer to decide when to set this command
        to the printer to print specific characters.

        :param cs: Character set variant.
        """
        self.current_character_set = cs
        return self._append_cmd('character_set', int_to_bytes(cs.value))

    def text(self, content: bytes | str | int, *, encoding=None) -> T:
        """Add text to the buffer.

        Make sure the string or bytes you send are compatible with the printer encoding.

        :param content:
        Text to add. Can be a string, bytes or integer.
        If it is a string, it will be encoded using the given encoding.
        If it is an integer, it will be converted to a single byte: 42 -> b'\x2a'.
        If it is bytes, it will be added as is.

        :param encoding:
        Encoding to use when converting a string to bytes.
        The natural encoding for ESC/P is `cp437`, but you can use any compatible encoding.
        Note that the ESC/P2 manual refers to 'PC437' encoding.
        """
        if encoding and not isinstance(encoding, str):
            raise ValueError('encoding valid only for a string')
        if not encoding:
            encoding = 'cp437'

        if isinstance(content, str):
            c = bytes(content, encoding)
            if len(c) != len(content):
                raise InvalidEncodingError(f'length of content encoded to {encoding} mismatches original length')
        elif isinstance(content, int):
            c = int_to_bytes(content)
        elif isinstance(content, bytes):
            c = content
        return self._append(c)

    def magic_text(self, content: str) -> T:
        """Print UTF-8 text using the magic encoding.

        This method will attempt to issue character set commands in order to print
        the given UTF-8 string.

        For instance, if the string contains an accented character 'é', it will switch to French
        character set (`CharacterSetVariant.FRANCE` i.e. ESC R \x01), print the character,
        and switch back to the original character set (USA is assumed).

        The switch occurs only if necessary, i.e. if the character is not available in the current character set.

        See `magic_encoding` for the mapping. Note that the mapping is far from complete.

        :param content: UTF-8 string to print.
        """
        initial_character_set = self.current_character_set
        for c in content:
            try:
                cs, code = magic_encoding[c]
                if self.current_character_set.value != cs.value:
                    self.character_set(cs)
                self.text(code)
            except KeyError:
                # no magic encoding needed
                if self.current_character_set.value != initial_character_set.value:
                    self.character_set(initial_character_set)
                self.text(c)

        if self.current_character_set != initial_character_set:
            self.character_set(initial_character_set)

        return self

    def cr_lf(self, how_many=1) -> T:
        return self._append(self._commands()['cr_lf'] * how_many)

    def bold(self, enabled: bool) -> T:
        return self._append_cmd('bold_on' if enabled else 'bold_off')

    def italic(self, enabled: bool) -> T:
        return self._append_cmd('italic_on' if enabled else 'italic_off')

    def double_strike(self, enabled: bool) -> T:
        """Prints each dot twice, with the second slightly below the first, creating bolder characters.

        On 9-pin printers, NLQ overrides double strike.
        """
        return self._append_cmd('double_strike_on' if enabled else 'double_strike_off')

    def underline(self, enabled: bool) -> T:
        """Turns on/off printing of a line below all characters and spaces"""
        return self._append_cmd('underline_on' if enabled else 'underline_off')

    def character_width(self, width: int) -> T:
        """Select the character width. This may set the point as well.

        :param width: 10, 12 or 15.
        On non- 9-pin printers, the point is set to 10.5 as well.
        """
        if width not in (10, 12, 15):
            raise ValueError(f'Invalid char width: ${width}')
        return self._append_cmd(f'character_width_{width}')

    def superscript(self, enabled: bool) -> T:
        """Prints characters in superscript mode.

        Prints characters that follow at about 2/3 their normal height
        """
        return self._append_cmd('superscript_on' if enabled else 'superscript_off')

    def subscript(self, enabled: bool) -> T:
        """Prints characters in subscript mode.

        Prints characters that follow at about 2/3 their normal height
        """
        return self._append_cmd('subscript_on' if enabled else 'subscript_off')

    def upper_control_codes_printing(self, enabled: bool) -> T:
        """Enable or disable printing of upper control codes.

        If enabled, tells the printer to treat codes from 128 to 159
        as printable characters instead of control code.
        """
        return self._append_cmd('upper_control_codes_printing_on' if enabled else 'upper_control_codes_printing_off')

    def control_codes_printing(self, enabled: bool) -> T:
        """Enable or disable printing of control codes.

        If enabled, tells the printer to treat codes 0–6, 16, 17, 21–23, 25, 26, 28–31, and 128–159 as
        printable character instead of control code.
        """
        return self._append_cmd('control_codes_printing', int_to_bytes(1 if enabled else 0))

    def typeface(self, tf: Typeface) -> T:
        return self._append_cmd('typeface', int_to_bytes(tf.value))

    def margin(self, margin: Margin, value: int) -> T:
        """Set a margin.

        The right margin value starts from the left margin.
        Example:
            left margin = 10 (1 inch left margin)
            right margin = 75 (1 inch right margin)

        Top margin is only available on ESC/P2 printers.

        Bottom margin on non ESC/P2 is only available with continuous paper.
        """
        if value < 0 or value > 255:
            raise ValueError(f'Invalid margin value: ${value}')
        if margin == Margin.TOP:
            raise NotImplementedError('Top margin is only available on ESC/P2 printers.')
        return self._append_cmd(f'margin_{margin.name.lower()}', int_to_bytes(value))

    @abstractmethod
    def line_spacing(self, numerator: int, denominator: int) -> T:
        """Set line spacing.

        Changing the line spacing after the page length does not affect the page length.
        Always set the line spacing before the page length.
        """
        raise NotImplementedError()

    def page_length(self, value: int, unit: PageLengthUnit) -> T:
        """Set page length.

        Always set the line spacing before the page length.
        Always set the page length before the paper is loaded
        or when the print position is at the top of the page.
        Setting the page length cancels the bottom margin.

        :param value The number of lines per page
        :param unit The unit of the page length
        """
        cmd = 'page_length_in_inches' if unit == PageLengthUnit.INCHES else 'page_length_in_lines'
        return self._append_cmd(cmd, int_to_bytes(value))

    def double_character_width(self, enabled: bool, one_line=False) -> T:
        """Select double-width printing.

        Doubles the width of all characters, spaces, and intercharacter spacing.
        """
        if one_line and enabled:
            return self._append_cmd('double_character_width_one_line')
        return self._append_cmd('double_character_width', int_to_bytes(1) if enabled else int_to_bytes(0))

    def double_character_height(self, enabled: bool) -> T:
        return self._append_cmd('double_character_height', int_to_bytes(1) if enabled else int_to_bytes(0))

    def extra_space(self, value: int) -> T:
        """Add extra space between characters.

        The fraction of inch depends on the number of pins.
        """
        if value < 0 or value > 255:
            raise ValueError(f'Invalid extra space value: ${value}')
        return self._append_cmd('extra_space', int_to_bytes(value))

    def condensed(self, enabled: bool) -> T:
        """Select condensed printing.

        1/17 inch if 10-cpi selected,
        1/20 inch if 12-cpi selected
        """
        return self._append_cmd('condensed_on' if enabled else 'condensed_off')

    def proportional(self, enabled: bool) -> T:
        """Select proportional printing.

        - Changes made to fixed-pitch printing are not effective until proportional printing is turned off.
        - Condensed printing is not effective when proportional printing is turned on.

        Printers not featuring this command:
        ActionPrinter Apex 80, ActionPrinter T-1000, ActionPrinter 2000, LX-400, LX-800, LX-810,
        LX-850, LX-1050
        """
        return self._append_cmd('proportional', int_to_bytes(1) if enabled else int_to_bytes(0))

    def justify(self, justification: Justification) -> T:
        """Set justification.

        - This is a non-recommended command as per Epson documentation,
          although no explanation is given.
        - Always set justification at the beginning of a line.
        - The printer performs full justification only if the width of the current line is greater than
          75% of the printing area width. If the line width is less than 75%, the printer left-justifies text.
        - You should not use commands that adjust the horizontal print position during full justification.
        - Justification is based on the font selected when the justification command is sent.
          Changing the font after setting justification can cause unpredictable results.
        """
        return self._append_cmd('justify', int_to_bytes(justification.value))

    def form_feed(self) -> T:
        return self._append_cmd('form_feed')

    def _append(self, b: bytes) -> T:
        self._buffer += bytearray(b)
        return self

    def _append_cmd(self, cmd: str, param: bytes = None) -> T:
        seq = self._commands()[cmd]
        if not seq:
            raise RuntimeError(f'No sequence found for {cmd}')
        if param:
            seq += param
        return self._append(seq)

    def clear(self) -> T:
        self._buffer = bytearray()
        return self

    @property
    def buffer(self) -> bytes:
        return bytes(self._buffer)


# Not implemented and reserved to 24/48 pin printers
# Outline + shadow (C-133)
# Scoring (C-127)
# Control paper loading/ejecting (C-157)



from typing import Self

from .commands import Commands, int_to_bytes, CharacterTable


class Commands_24_48_Pin(Commands):

    specific_cmds = {
        'line_spacing_n_180': b'\x1b3',
        'line_spacing_n_360': b'\x1b+',
    }

    def _commands(self):
        cmds = super()._commands().copy()
        cmds.update(self.specific_cmds)
        return cmds

    def is_valid_character_table(self, table: int) -> bool:
        return 0 <= table <= 3

    def select_character_table(self, table: int) -> Self:
        return super().select_character_table(table)

    def assign_character_table(self, table: int, ct: CharacterTable) -> Self:
        return super().assign_character_table(table, ct)

    def extra_space(self, value: int) -> Self:
        """Add extra space between characters.

        Add n/180 inch on 24/48-pin printers in LQ mode,
        n/120 inch on 24/48-pin printers in Draft mode.
        """
        return super().extra_space(value)

    def line_spacing(self, numerator: int, denominator: int) -> Self:
        """Set line spacing.

        Changing the line spacing after the page length does not affect the page length.
        Always set the line spacing before the page length.

        Valid combinations: 1/6, 1/8, n/180, n/360
        """
        match numerator, denominator:
            case 1, 6:
                return self._append_cmd('line_spacing_1_6')
            case 1, 8:
                return self._append_cmd('line_spacing_1_8')
            case n, 180:
                return self._append_cmd('line_spacing_n_180', int_to_bytes(n))
            case n, 360:
                return self._append_cmd('line_spacing_n_360', int_to_bytes(n))
            case _:
                raise ValueError(f'Invalid line spacing: {numerator}/{denominator}')



from src.escp import Commands
from src.escp.commands.commands import T


class CommandsDefault(Commands):
    def line_spacing(self, numerator: int, denominator: int) -> T:
        raise NotImplementedError()

    def is_valid_character_table(self, table: int) -> bool:
        pass



from typing import Self

from escp.commands.commands import T

from .commands import Commands, int_to_bytes, CharacterTable


class Commands_9_Pin(Commands):

    specific_cmds = {
        'line_spacing_n_216': b'\x1b3',
    }

    def _commands(self):
        cmds = super()._commands().copy()
        cmds.update(self.specific_cmds)
        return cmds

    def is_valid_character_table(self, table: int) -> bool:
        return 0 <= table <= 1

    def select_character_table(self, table: int) -> Self:
        """Select the character table.

        0: Italic
        1: PC437(US) ('cp437' in Python)
        [C-79].

        :param table: Character table number.
        """
        return super().select_character_table(table)

    def assign_character_table(self, table: int, ct: CharacterTable) -> T:
        return super().assign_character_table(table, ct)

    def extra_space(self, value: int) -> Self:
        """Add extra space between characters.

        Add n/120 inch on 9-pin printers.
        """
        return super().extra_space(value)

    def line_spacing(self, numerator: int, denominator: int) -> Self:
        """Set line spacing.

        Changing the line spacing after the page length does not affect the page length.
        Always set the line spacing before the page length.

        Valid combinations: 1/6, 1/8, n/216
        """
        match numerator, denominator:
            case 1, 6:
                return self._append_cmd('line_spacing_1_6')
            case 1, 8:
                return self._append_cmd('line_spacing_1_8')
            case n, 216:
                return self._append_cmd('line_spacing_n_216', int_to_bytes(n))
            case _:
                raise ValueError(f'Invalid line spacing: {numerator}/{denominator}')




import pytest
from src.escp import InvalidEncodingError

from test import CommandsDefault


@pytest.fixture
def commands():
    return CommandsDefault()


@pytest.mark.parametrize('encoding,', ['cp437', 'utf-8'])
def test_encoding_independent(commands, encoding):
    assert commands.text("Hello world", encoding=encoding).buffer == b'Hello world'


usa_char_set_fixture = [
    ('#', b'\x23'), ('$', b'\x24'), ('@', b'\x40'), ('[', b'\x5b'),
    ('\\', b'\x5c'), (']', b'\x5d'), ('^', b'\x5e'), ('`', b'\x60'),
    ('{', b'\x7b'), ('|', b'\x7c'), ('}', b'\x7d'), ('~', b'\x7e')
]


@pytest.mark.parametrize('char,expected', usa_char_set_fixture)
def test_cp437_usa(commands, char, expected):
    assert commands.text(char, encoding='cp437').buffer == expected


@pytest.mark.parametrize('char,expected', usa_char_set_fixture)
def test_utf8_usa(commands, char, expected):
    assert commands.text(char, encoding='utf-8').buffer == expected


def test_utf8_french(commands):
    # should send \x7b with international character set 1 (France)
    with pytest.raises(InvalidEncodingError):
        commands.text("é", encoding='utf-8')


def test_utf8_emoji(commands):
    # Can never be encoded in cp437
    with pytest.raises(InvalidEncodingError):
        commands.text("😢", encoding='utf-8')


def test_cp437_emoji(commands):
    # Can never be encoded in cp437
    with pytest.raises(UnicodeError):
        commands.text("😢", encoding='cp437')



import pytest

from src.escp import Commands_9_Pin, CharacterTable


@pytest.fixture
def commands():
    return Commands_9_Pin()


@pytest.mark.parametrize('numerator,denominator,codes', [(1, 6, b'\x1b2'), (1, 8, b'\x1b0'), (45, 216, b'\x1b3\x2d')])
def test_line_spacing(commands, numerator, denominator, codes):
    assert commands.line_spacing(numerator, denominator).buffer == codes


@pytest.mark.parametrize('ct,expected_buffer', [(0, b'\x1bt\x00'), (1, b'\x1bt\x01')])
def test_select_character_table(commands, ct, expected_buffer):
    assert commands.select_character_table(ct).buffer == expected_buffer


def test_select_character_table_invalid(commands):
    with pytest.raises(ValueError):
        commands.select_character_table(2)


@pytest.mark.parametrize(
    'table,character_table,expected_buffer',
    [
        (0, CharacterTable.ITALIC, b'\x1b\x28\x74\x03\x00\x00\x00\x00'),
        (0, CharacterTable.PC437_US, b'\x1b\x28\x74\x03\x00\x00\x01\x00'),
        (1, CharacterTable.PC437_US, b'\x1b\x28\x74\x03\x00\x01\x01\x00'),
        (1, CharacterTable.PC_AR864, b'\x1b\x28\x74\x03\x00\x01\x0d\x20'),
    ]
)
def test_assign_character_table(commands, table, character_table, expected_buffer):
    assert commands.assign_character_table(table, character_table).buffer == expected_buffer



from ..commands import Commands, Justification, PageLengthUnit, CharacterTable
from ..printer import Printer
from .common import print_and_reset


PAGE_LENGTH_INCHES = 12


def print_char_table(printers: [Printer], cmd: Commands):
    def _print_and_reset():
        print_and_reset(printers, cmd)

    cmd.init()

    cmd.text('CHARACTER TABLES').cr_lf()
    cmd.text('================').cr_lf(3)
    cmd.text('Samples: 140-149, 160-169').cr_lf(2)

    for ct in CharacterTable:
        cmd.init().text('# ').text(str(ct)).cr_lf()
        cmd.assign_character_table(1, ct)
        ranges = [range(140, 150), range(160, 170)]
        for rrange in ranges:
            for char in rrange:
                cmd.text(char)
            cmd.text('  ')
        cmd.cr_lf(2)
    cmd.form_feed()

    _print_and_reset()

    for printer in printers:
        printer.close()



from ..commands import Commands, Justification, Margin, Typeface
from ..printer import Printer

from .common import print_and_reset


def fox() -> bytes:
    return b'The quick brown fox jumps over the lazy dog'


def lorem() -> bytes:
    return (
        b'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor\x0d\x0a'
        b'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,\x0d\x0a'
        b'quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.\x0d\x0a'
        b'Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu\x0d\x0a'
        b'fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in\x0d\x0a'
        b'culpa qui officia deserunt mollit anim id est laborum.'
    )


def mini_lorem() -> bytes:
    return (
        b'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor\x0d\x0a'
        b'incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, ...'
    )


def sep_line(count=20) -> bytes:
    return b'-' * count


def print_test_page(printers: [Printer], cmd: Commands):
    def _next_sequence():
        cmd.clear().init().draft(False).typeface(Typeface.SANS_SERIF)

    def _print_and_reset(prepare_next_sequence=True):
        print_and_reset(printers, cmd, _next_sequence if prepare_next_sequence else None)

    # Init
    _print_and_reset()

    # Hello
    cmd.text('ESC/P direct printing test page').cr_lf(2)
    _print_and_reset()

    # Text enhancement
    cmd.text('Text enhancements').cr_lf()
    cmd.text('Bold').cr_lf()
    cmd.bold(True).text(fox()).bold(False).cr_lf()
    cmd.text('Italic').cr_lf()
    cmd.italic(True).text(fox()).italic(False).cr_lf()
    cmd.text('Double strike').cr_lf()
    cmd.double_strike(True).text(fox()).double_strike(False).cr_lf()
    cmd.text('Underline').cr_lf()
    cmd.underline(True).text(fox()).underline(False).cr_lf()
    cmd.cr_lf()
    _print_and_reset()

    # Char width
    cmd.text('Character width').cr_lf()
    for width in [10, 12, 15]:
        cmd \
            .text(f'1/{width} char width') \
            .cr_lf() \
            .character_width(width) \
            .text(fox()) \
            .cr_lf()
    cmd.cr_lf()
    _print_and_reset()

    # Super/subscript
    cmd.text('Superscript').cr_lf()
    cmd.text('Some reference').superscript(True).text('[1]').superscript(False).cr_lf()
    cmd.text('Water is H').subscript(True).text('2').subscript(False).text('O').cr_lf()
    cmd.cr_lf()
    _print_and_reset()

    # Typeface
    cmd.text('Typeface').cr_lf()
    cmd.text('Roman').cr_lf()
    cmd.typeface(Typeface.ROMAN)
    cmd.text('    ').text(fox()).cr_lf()
    cmd.typeface(Typeface.SANS_SERIF)
    cmd.text('Sans Serif').cr_lf()
    cmd.text('    ').text(fox()).cr_lf()
    cmd.cr_lf()
    _print_and_reset()

    # Margins (left)
    cmd.text('Margins (left)')
    for margin in [0, 4, 8]:
        cmd \
            .margin(Margin.LEFT, margin) \
            .text(f'[x] text started at col {margin}') \
            .cr_lf()
    cmd.cr_lf()
    _print_and_reset()

    # Character size
    cmd.text('Character size').cr_lf()
    cmd.double_character_width(True).text('Double character width').double_character_width(False).cr_lf(2)
    cmd.double_character_height(True).text('Double character height').double_character_height(False).cr_lf(2)
    cmd \
        .double_character_width(True) \
        .double_character_height(True) \
        .text('Double character width and height') \
        .double_character_width(False) \
        .double_character_height(False) \
        .cr_lf(2)
    _print_and_reset()

    # Character spacing
    cmd.text('Extra space between characters').cr_lf()
    for extra_space in [1, 5, 10]:
        cmd \
            .text(f'{extra_space}/120"') \
            .cr_lf() \
            .extra_space(extra_space) \
            .text(fox()) \
            .cr_lf()
        _print_and_reset()
    cmd.cr_lf()

    # Condensed
    cmd.text('Condensed text').cr_lf()
    cmd.condensed(True).text(fox()).text('. ').text(fox()).condensed(False).cr_lf(2)
    _print_and_reset()

    # Line spacing
    cmd \
        .text('Line spacing').cr_lf() \
        .text('(not specified)').cr_lf() \
        .text(fox()).cr_lf().text(fox()).cr_lf() \
        .text('1/8').cr_lf() \
        .line_spacing(1, 8) \
        .text(fox()).cr_lf().text(fox()).cr_lf() \
        .text('1/6').cr_lf() \
        .line_spacing(1, 6) \
        .text(fox()).cr_lf().text(fox()).cr_lf() \
        .cr_lf()
    _print_and_reset()

    # Proportional
    cmd.text('Proportional text').cr_lf()
    cmd.proportional(True).text(lorem()).proportional(False).cr_lf(2)
    _print_and_reset()

    # Justification
    cmd.text('Justification (with proportional)').cr_lf()
    cmd.proportional(True)
    cmd.justify(Justification.LEFT).text(fox()).cr_lf()
    cmd.justify(Justification.CENTER).text(fox()).cr_lf()
    cmd.justify(Justification.RIGHT).text(fox()).cr_lf(2)
    _print_and_reset()

    cmd.proportional(True)
    cmd.justify(Justification.CENTER).text(lorem()).cr_lf(2)
    _print_and_reset()

    cmd.proportional(True)
    cmd.justify(Justification.FULL).text(lorem()).cr_lf(2)
    cmd.proportional(False)
    _print_and_reset()

    # Printable control codes
    cmd.text('Printable control codes - upper control codes').cr_lf()
    for i in range(128, 160):
        cmd \
            .text(str(i)) \
            .text(' ') \
            .upper_control_codes_printing(True) \
            .text(i) \
            .upper_control_codes_printing(False) \
            .text('   ')
    cmd.cr_lf()
    _print_and_reset()

    cmd.form_feed()
    _print_and_reset(prepare_next_sequence=False)

    for printer in printers:
        printer.close()



import sys

from .printer import Printer


class OutputPrinter(Printer):

    def __init__(self, output=sys.stdout):
        self.output = output

    def send(self, sequence: bytes):
        self.output.write(str(sequence))
        self.output.flush()

    def close(self):
        close_method = getattr(self.output, 'close', None)
        if callable(close_method):
            close_method()



from ..commands import Commands, Justification, PageLengthUnit
from ..printer import Printer
from .common import print_and_reset


PAGE_LENGTH_INCHES = 12


def print_poem(printers: [Printer], cmd: Commands):
    def _print_and_reset():
        print_and_reset(printers, cmd)

    text = """When I heard the learn'd astronomer
When the proofs, the figures, were ranged in columns before me
When I was shown the charts and diagrams, to add, divide, and measure them 
When I sitting heard the astronomer where he lectured
with much applause in the lecture-room
How soon unaccountable I became tired and sick
Till rising and gliding out I wander'd off by myself
In the mystical moist night-air, and from time to time
Look'd up in perfect silence at the stars
"""
    cmd \
        .init() \
        .page_length(PAGE_LENGTH_INCHES, PageLengthUnit.INCHES) \
        .justify(Justification.CENTER) \
        .proportional(True) \
        .line_spacing(45, 216) \
        .bold(True).text('When I heard the learn\'d astronomer').bold(False).cr_lf(2) \
        .italic(True).text('by Walt Whitman').italic(False).cr_lf(2) \
        .text(text) \
        .form_feed()
    _print_and_reset()

    for printer in printers:
        printer.close()



import pytest
from src.escp import Margin, PageLengthUnit, Typeface

from test import CommandsDefault


@pytest.fixture
def commands():
    return CommandsDefault()


def test_draft_mode_on(commands):
    assert commands.draft(True).buffer == b'\x1bx\x01'


def test_draft_mode_off(commands):
    assert commands.draft(False).buffer == b'\x1bx\x00'


def test_text_str(commands):
    assert commands.text("Hello world").buffer == b'Hello world'


def test_text_bytes(commands):
    assert commands.text(b"Hello world").buffer == b'Hello world'


def test_text_int(commands):
    assert commands.text(42).buffer == b'*'


def test_cr_lf(commands):
    assert commands.cr_lf().buffer == b'\x0d\x0a'


def test_cr_lf_several(commands):
    assert commands.cr_lf(2).buffer == b'\x0d\x0a\x0d\x0a'


def test_bold_on(commands):
    assert commands.bold(True).buffer == b'\x1bE'


def test_bold_off(commands):
    assert commands.bold(False).buffer == b'\x1bF'


def test_double_strike_on(commands):
    assert commands.double_strike(True).buffer == b'\x1bG'


def test_double_strike_off(commands):
    assert commands.double_strike(False).buffer == b'\x1bH'


def test_italic_on(commands):
    assert commands.italic(True).buffer == b'\x1b4'


def test_italic_off(commands):
    assert commands.italic(False).buffer == b'\x1b5'


def test_underline_on(commands):
    assert commands.underline(True).buffer == b'\x1b-\x01'


def test_underline_off(commands):
    assert commands.underline(False).buffer == b'\x1b-\x00'


@pytest.mark.parametrize("width,byte_param", [(10, b'\x1bP'), (12, b'\x1bM'), (15, b'\x1bg')])
def test_character_width(commands, width, byte_param):
    assert commands.character_width(width).buffer == byte_param


def test_character_width_invalid(commands):
    with pytest.raises(ValueError):
        commands.character_width(11)


def test_superscript_on(commands):
    assert commands.superscript(True).buffer == b'\x1bS\x00'


def test_superscript_off(commands):
    assert commands.superscript(False).buffer == b'\x1bT'


def test_subscript_on(commands):
    assert commands.subscript(True).buffer == b'\x1bS\x01'


def test_subscript_off(commands):
    assert commands.subscript(False).buffer == b'\x1bT'


def test_upper_control_codes_printing_on(commands):
    assert commands.upper_control_codes_printing(True).buffer == b'\x1b\x36'


def test_upper_control_codes_printing_off(commands):
    assert commands.upper_control_codes_printing(False).buffer == b'\x1b\x37'


def test_control_codes_printing_on(commands):
    assert commands.control_codes_printing(True).buffer == b'\x1b\x49\x01'


def test_control_codes_printing_off(commands):
    assert commands.control_codes_printing(False).buffer == b'\x1b\x49\x00'


@pytest.mark.parametrize('typeface,value', [(Typeface.ROMAN, b'\x1bk\x00'), (Typeface.SANS_SERIF, b'\x1bk\x01')])
def test_typeface(commands, typeface, value):
    assert commands.typeface(typeface).buffer == value


@pytest.mark.parametrize(
    'side,codes', [
        (Margin.RIGHT, b'\x1bQ\x05'),
        (Margin.BOTTOM, b'\x1bN\x05'),
        (Margin.LEFT, b'\x1bl\x05'),
    ]
)
def test_margin(commands, side, codes):
    assert commands.margin(side, 5).buffer == codes


def test_margin_invalid(commands):
    with pytest.raises(NotImplementedError):
        commands.margin(Margin.TOP, 5)


def test_line_spacing(commands):
    with pytest.raises(NotImplementedError):
        commands.line_spacing(1, 1)


def test_page_length_in_lines(commands):
    assert commands.page_length(20, PageLengthUnit.LINES).buffer == b'\x1b\x43\x14'


def test_page_length_in_lines(commands):
    assert commands.page_length(20, PageLengthUnit.LINES).buffer == b'\x1b\x43\x14'


def test_page_length_in_inch(commands):
    assert commands.page_length(20, PageLengthUnit.INCHES).buffer == b'\x1bC\x00\x14'


def test_clear(commands):
    commands.text("Hello world")
    assert commands.clear().buffer == b''



from .parameters import CharacterSetVariant


# International character set [R-41]
# Very basic table, far from complete
magic_encoding = {
    "#": (CharacterSetVariant.USA, b'\x23'),
    "$": (CharacterSetVariant.USA, b'\x24'),
    "@": (CharacterSetVariant.USA, b'\x40'),
    "[": (CharacterSetVariant.USA, b'\x5b'),
    "\\": (CharacterSetVariant.USA, b'\x5c'),
    "]": (CharacterSetVariant.USA, b'\x5d'),
    "^": (CharacterSetVariant.USA, b'\x5e'),
    "`": (CharacterSetVariant.USA, b'\x60'),
    "{": (CharacterSetVariant.USA, b'\x7b'),
    "|": (CharacterSetVariant.USA, b'\x7c'),
    "}": (CharacterSetVariant.USA, b'\x7d'),
    "˜": (CharacterSetVariant.USA, b'\x7e'),
    "à": (CharacterSetVariant.FRANCE, b'\x40'),
    "°": (CharacterSetVariant.FRANCE, b'\x5b'),
    "ç": (CharacterSetVariant.FRANCE, b'\x5c'),
    "§": (CharacterSetVariant.FRANCE, b'\x5d'),
    "é": (CharacterSetVariant.FRANCE, b'\x7b'),
    "ù": (CharacterSetVariant.FRANCE, b'\x7c'),
    "è": (CharacterSetVariant.FRANCE, b'\x7d'),
    "¨": (CharacterSetVariant.FRANCE, b'\x7e'),
    "©": (CharacterSetVariant.LEGAL, b'\x7b'),
    "®": (CharacterSetVariant.LEGAL, b'\x7c'),
}
